#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2022/02/09 11:14


"""
Apache ECharts: https://echarts.apache.org/zh/index.html
python版：pyechats的封装
"""

from pyxllib.prog.pupil import check_install_package

check_install_package('pyecharts')

# import types

import pyecharts
from pyecharts import options as opts
from pyecharts.commons.utils import JsCode
from pyecharts.globals import ChartType
from pyecharts import types
from pyecharts.charts import Bar, Line, Radar
from pyecharts.charts.chart import Chart

from pyxllib.prog.pupil import inject_members
from pyxllib.prog.specialist import TicToc
from pyxllib.file.specialist import XlPath


class XlChart(Chart):
    def set_title(self, title):
        self.set_global_opts(title_opts=pyecharts.options.TitleOpts(title=title))

    def add_series(self, name, data, *, type=None, color=None, labels=None,
                   **kwargs):
        """ 垃圾pyecharts，毁我青春~~

        很多图x都不是等间距的，pyecharts处理不等间距x好像有很大的不兼容问题
        需要手动添加x、y坐标位置

        :param list|tuple labels: 直接提供现成的文本标签

        """
        if type is None:
            type = self.__class__.__name__.lower()

        if labels:
            s = [[x[0][0], x[1]] for x in zip(data, labels)]
            fmt = JsCode(f"function(x){{var m = new Map({s}); return m.get(x.value[0]);}}")
            if 'label' not in kwargs:
                kwargs['label'] = opts.LabelOpts(is_show=True, formatter=fmt)
            elif isinstance(kwargs['label'], opts.LabelOpts):
                kwargs['label'].opts['formatter'] = fmt
            else:
                kwargs['label']['show'] = True
                kwargs['label']['formatter'] = fmt

        self._append_color(color)
        # self._append_legend(name, is_selected=True)
        self._append_legend(name)

        self.options.get('series').append(
            {
                'type': type,
                'name': name,
                'data': data,
                **kwargs,
            }
        )
        return self


inject_members(XlChart, Chart)


class XlLine(Line):
    @classmethod
    def from_dict(cls, yaxis, xaxis=None, *, title=None,
                  xaxis_name=None, yaxis_name=None, **kwargs):
        """ 查看一个数据的折线图

        :param list|dict yaxis: 列表数据，或者字典name: values表示的多组数据
        """
        c = cls()

        if isinstance(yaxis, (list, tuple)):
            yaxis = {'value': yaxis}
        if xaxis is None:
            xaxis = list(range(max([len(v) for v in yaxis.values()])))
        c.add_xaxis(xaxis)
        for k, v in yaxis.items():
            c.add_yaxis(k, v)

        configs = {
            'tooltip_opts': opts.TooltipOpts(trigger="axis"),
            **kwargs,
        }
        c.set_global_opts(tooltip_opts=opts.TooltipOpts(trigger="axis"))
        if title:
            configs['title_opts'] = opts.TitleOpts(title=title)
        if xaxis_name:
            configs['xaxis_opts'] = opts.AxisOpts(name=xaxis_name, axislabel_opts=opts.LabelOpts(rotate=45, interval=0))
        if yaxis_name:
            configs['yaxis_opts'] = opts.AxisOpts(name=yaxis_name)
        c.set_global_opts(**configs)
        return c


class XlBar(Bar):

    @classmethod
    def from_dict(cls, yaxis, xaxis=None, *, title=None,
                  xaxis_name=None, yaxis_name=None, **kwargs):
        """ 查看一个数据的条形图

        :param list|dict yaxis: 列表数据，或者字典name: values表示的多组数据
        """
        c = cls()

        if isinstance(yaxis, (list, tuple)):
            yaxis = {'value': yaxis}
        if xaxis is None:
            xaxis = list(range(max([len(v) for v in yaxis.values()])))
        c.add_xaxis(xaxis)
        for k, v in yaxis.items():
            c.add_yaxis(k, v)

        configs = {
            'tooltip_opts': opts.TooltipOpts(trigger="axis"),
            **kwargs,
        }
        c.set_global_opts(tooltip_opts=opts.TooltipOpts(trigger="axis"))
        if title:
            configs['title_opts'] = opts.TitleOpts(title=title)
        if xaxis_name:
            configs['xaxis_opts'] = opts.AxisOpts(name=xaxis_name)
        if yaxis_name:
            configs['yaxis_opts'] = opts.AxisOpts(name=yaxis_name)
        c.set_global_opts(**configs)
        return c

    @classmethod
    def from_list(cls, yaxis, xaxis=None, *, title=None):
        """
        >> browser(pyecharts.charts.Bar.from_list(numbers))
        """
        return cls.from_dict({'value': list(yaxis)}, xaxis=xaxis, title=title)

    @classmethod
    def from_data_split_into_groups(cls, data, groups, *, title=None):
        """根据给定的组数自动拆分数据并生成条形图
        :param list data: 数据清单
        :param int groups: 要拆分成的组数
        """
        # 找到最大值和最小值
        min_val, max_val = min(data), max(data)

        # 计算间隔
        interval = (max_val - min_val) / groups

        # 分组和标签
        group_counts = [0] * groups
        labels = []
        # todo 如果数据量特别大，这里应该排序后，再用特殊方法计算分组
        for value in data:
            index = min(int((value - min_val) / interval), groups - 1)
            group_counts[index] += 1

        for i in range(groups):
            labels.append(f"{min_val + interval * i:.2f}-{min_val + interval * (i + 1):.2f}")
        # t = cls.from_dict({'value': group_counts}, xaxis=labels, title=title)

        return cls.from_dict({'value': group_counts}, xaxis=labels, title=title)


class XlRadar(Radar):
    """ 雷达图 """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.color_idx = 0

    def add(
            self,
            series_name: str,
            data: types.Sequence[types.Union[opts.RadarItem, dict]],
            *,
            label_opts=None,
            color: types.Optional[str] = None,
            linestyle_opts=None,
            **kwargs
    ):
        """ 标准库(2.0.5版)的雷达图颜色渲染有问题，这里要增加一个修正过程 """
        if label_opts is None:
            label_opts = opts.LabelOpts(is_show=False)

        if linestyle_opts is None:
            linestyle_opts = opts.LineStyleOpts(color=self.colors[self.color_idx % len(self.colors)])
            self.color_idx += 1
        elif linestyle_opts.get('color') is None:
            linestyle_opts.update(color=self.colors[self.color_idx % len(self.colors)])
            self.color_idx += 1

        if color is None:
            color = linestyle_opts.get('color')

        return super(XlRadar, self).add(series_name, data,
                                        label_opts=label_opts,
                                        color=color,
                                        linestyle_opts=linestyle_opts,
                                        **kwargs)


inject_members(XlBar, Bar)


def render_echart_html(title='Awesome-pyecharts', body=''):
    from pyxllib.text.xmllib import get_jinja_template
    return get_jinja_template('echart_base.html').render(title=title, body=body)


# 绘制帕累托累计图
def draw_pareto_chart(data, accuracy=0.1, *, title='帕累托累积权重', value_unit_type='K'):
    from pyxllib.algo.stat import pareto_accumulate
    pts, labels = pareto_accumulate(data, accuracy=accuracy, value_unit_type=value_unit_type)
    x = Line()
    x.add_series(title, pts, labels=labels, label={'position': 'right'})
    x.set_global_opts(
        # x轴末尾要故意撑大一些，不然有部分内容会显示不全
        xaxis_opts=opts.AxisOpts(name='条目数', max_=int(float(f'{pts[-1][0] * 1.2:.2g}'))),
        yaxis_opts=opts.AxisOpts(name='累积和')
    )
    return x


if __name__ == '__main__':
    with TicToc(__name__):
        pass



from json import dumps as jsonDumps
from typing import List, Tuple, Union, Literal
from openai import OpenAI
from .chat import system_msg, user_msg, assistant_msg, Chat, AKPool


class GroupChat(Chat):

    roles = {
        "苏轼":'宋朝诗人，他的词风格独特，既有儒家的教诲，又有生活的乐趣。',
        "李清照":'宋代著名的女词人，其词句优美，情感真挚。',
        "杜甫":'唐朝著名诗人。',
        "小许":"一个聪明的程序员",
        "小郑":"一个帅气的男人",
        "小张":"一个漂亮的女人"
    }

    _top_messages: List[dict] = None

    def __init__(self, *vs, **kvs):
        Chat.__init__(self, *vs, **kvs)
        self.roles = self.roles.copy()
        self._new_dialogs = []
        if not self.__class__._top_messages:
            self.__class__._top_messages = self._render_top_messages()
    
    # 欺骗编辑器进行代码提示
    if len(__file__) < 0:
        def __init__(self,
                    # kwargs
                    api_key: Union[str, AKPool],
                    base_url: str = None,  # base_url 参数用于修改基础URL
                    timeout=None,
                    max_retries=None,
                    http_client=None,
                    # request_kwargs
                    model: Literal["gpt-4-1106-preview", "gpt-4-vision-preview", "gpt-4", "gpt-4-0314", "gpt-4-0613",
                                    "gpt-4-32k", "gpt-4-32k-0314", "gpt-4-32k-0613", "gpt-3.5-turbo"] = "gpt-3.5-turbo",
                    # Chat
                    msg_max_count=None,
                    # kwargs
                    **kwargs,
                    ):
            ...
    
    def add_dialog(self, speaker:str, audiences:list, remark:str):
        '''
        add_dialog('李白', ['杜甫', '小许'], '你们好呀')
        add_dialog('杜甫', ['李白'], '你好, 你今天写诗了吗?')
        '''
        self.roles.setdefault(speaker, '')
        for x in audiences:
            self.roles.setdefault(x, '')
        self._new_dialogs.append( f"【{speaker}】对【{'、'.join(audiences)}】说：{remark}".replace('\n', ' ') )
    
    def _get_user_text(self, roles: List[Tuple[str, List[str]]]):
        # 补充人物数据
        for speaker, audiences in roles:
            self.roles.setdefault(speaker, '')
            for x in audiences:
                self.roles.setdefault(x, '')
        # 新增的对话记录
        _new_dialogs = [f"{i+1}、{x}" for i, x in enumerate(self._new_dialogs)]
        _new_dialogs = '\n'.join(_new_dialogs)
        # 需要模拟的对话
        roles = [{'speaker':speaker, 'audiences':'、'.join(audiences), 'remark':''} for speaker, audiences in roles]
        roles = jsonDumps(roles, ensure_ascii=False)
        # 用户的发言
        text = f'''以下是新增的对话记录：\n\n{_new_dialogs}\n\n请根据以下JSON文档的内容，模拟角色的视角进行发言，并把发言填充到'remark'字段中：\n\n```json\n{roles}\n```\n\n请把填充完整后的JSON文档发给我，只返回JSON文档即可，勿包含任何其它信息，否则会干扰我的解析。'''
        return text
    
    def request(self, roles: List[Tuple[str, List[str]]], **kwargs):
        '''
        request([
            ('苏轼', ['李清照', '杜甫']),
            ('小明', ['小东']),
        ])
        '''
        messages = [{"role": "user", "content": self._get_user_text(roles)}]
        self.recently_request_data = {
            'api_key': (api_key := self._akpool.fetch_key()),
        }
        completion = OpenAI(api_key=api_key, **self._kwargs).chat.completions.create(**{
            **self._request_kwargs,  # 全局参数
            **kwargs,  # 单次请求的参数覆盖全局参数
            "messages": self._top_messages + list(self._messages) + messages,
            "stream": False,
        })
        answer: str = completion.choices[0].message.content
        self._messages.add_many(*messages, {"role": "assistant", "content": answer})
        self._new_dialogs = []
        return answer
    
    def _render_top_messages(self):
        messages = []
        # 任务提示
        messages.append(system_msg(f'''以下的JSON文档描述了一些人物的角色信息：\n\n```json\n{jsonDumps(self.roles, ensure_ascii=False)}\n```\n\n你需要了解这些人物的信息。user每次都会收集这些人物的对话记录发送给你，并让你站在其中的某些角色的视角上对其它角色进行模拟发言，你每次返回JSON文档即可，勿包含任何其它信息，否则会干扰user的解析。你在发言时可以编造，比如在回答年龄时，可以随意编一个年龄，但要注意上下文逻辑闭环。'''))
        # 第一轮教学
        self.add_dialog('李白', ['杜甫', '小许'], '你们好呀')
        self.add_dialog('杜甫', ['李白'], '你好, 你今天写诗了吗?')
        self.add_dialog('小许', ['李白'], '你好, 你吃饭了吗?')
        messages.append(user_msg(self._get_user_text([('李白',['小许']), ('李白',['杜甫']), ('李白', ['杜甫','小许'])])))
        messages.append(assistant_msg('''[{"speaker": "李白", "audiences": "小许", "remark": "我今天写诗了"}, {"speaker": "李白", "audiences": "杜甫", "remark": "我吃饭了"}, {"speaker": "李白", "audiences": "杜甫、小许", "remark": "你们有什么有趣的事情分享吗?"}]'''))
        self._new_dialogs = []
        # 第二轮教学
        self.add_dialog('小郑', ['小张'], '你是谁?')
        self.add_dialog('小张', ['小郑'], '我叫小张,今年13岁')
        self.add_dialog('小许', ['小郑', '小张'], '你们是哪里人?')
        messages.append(user_msg(self._get_user_text([('小郑',['小张']), ('小郑',['小许']), ('小张', ['小许'])])))
        messages.append(assistant_msg('''[{"speaker": "小郑", "audiences": "小张", "remark": "哦哦, 我比你大1岁"}, {"speaker": "小郑", "audiences": "小许", "remark": "我是河南人"}, {"speaker": "小张", "audiences": "小许", "remark": "不告诉你"}]'''))
        self._new_dialogs = []
        # 第三轮教学
        self.add_dialog('小许', ['小张'], '呵呵,这么神秘呀?')
        self.add_dialog('小许', ['小郑'], '你知道小张是哪里人吗?')
        messages.append(user_msg(self._get_user_text([('小郑',['小许']), ('小张',['小许'])])))
        messages.append(assistant_msg('''[{"speaker": "小郑", "audiences": "小许", "remark": "哈哈, 我也不知道"}, {"speaker": "小张", "audiences": "小许", "remark": "哈哈, 开个玩笑啦, 我是湖北人"}]'''))
        self._new_dialogs = []
        # 导出
        messages = [dict(x) for x in messages]
        return messages


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2020/06/06

import sys
from collections import defaultdict
import json
import os
import time
import random

import mss
import numpy as np
from pandas.api.types import is_list_like

if sys.platform == 'win32':
    import pyautogui
    import win32gui

try:
    import pyscreeze  # NOQA pyautogui安装的时候会自动安装依赖的pyscreeze
except ModuleNotFoundError:
    pass

from pyxllib.prog.newbie import first_nonnone, round_int
from pyxllib.prog.pupil import xlwait, DictTool, check_install_package
from pyxllib.prog.specialist import TicToc
from pyxllib.algo.geo import ComputeIou, ltrb2xywh, xywh2ltrb, ltrb2polygon
from pyxllib.algo.shapelylib import ShapelyPolygon
from pyxllib.file.specialist import File, Dir, XlPath
from pyxllib.cv.expert import xlcv, xlpil
from pyxlpr.data.labelme import LabelmeDict


def __20210531():
    """ 以前设想的一个跟labelme结合的自动化操作类 """


class AutoGuiLabelData:
    """ AutoGuiLabelData """

    def __init__(self, root):
        """
        data：dict
            key: 相对路径/图片stem （这一节称为loc）
                label名称：对应属性dict1 （跟loc拼在一起，称为loclabel）
                label2：dict2
        """
        self.root = XlPath(root)
        self.data = defaultdict(dict)
        self.imfiles = {}

        for file in self.root.rglob_files('*.json'):
            lmdict = file.read_json()
            imfile = file.with_name(lmdict['imagePath'])
            img = xlcv.read(imfile)
            loc = str(file.relpath(self.root)).replace('\\', '/')[:-5]
            self.imfiles[loc] = imfile

            for shape in lmdict['shapes']:
                attrs = self.parse_shape(shape, img)
                self.data[loc][attrs['text']] = attrs

    @classmethod
    def parse_shape(cls, shape, image=None):
        """ 解析一个shape的数据为dict字典，会把一些主要的几何特征也加进label字典字段中 """
        # 1 解析原label为字典
        # 如果是来自普通的label，会把原文本自动转换为text字段。如果是来自xllabelme，则一般默认就是字典值，带text字段。
        attrs = DictTool.json_loads(shape['label'], 'text')
        attrs.update(DictTool.sub(shape, ['label']))

        # 2 中心点 center
        shape_type = shape['shape_type']
        pts = shape['points']
        if shape_type in ('rectangle', 'polygon', 'line'):
            attrs['center'] = np.array(np.array(pts).mean(axis=0), dtype=int).tolist()
        elif shape_type == 'circle':
            attrs['center'] = pts[0]

        # 3 外接矩形 rect
        if shape_type in ('rectangle', 'polygon'):
            attrs['ltrb'] = np.array(pts, dtype=int).reshape(-1).tolist()
        elif shape_type == 'circle':
            x, y = pts[0]
            r = ((x - pts[1][0]) ** 2 + (y - pts[1][1]) ** 2) ** 0.5
            attrs['ltrb'] = [round_int(v) for v in [x - r, y - r, x + r, y + r]]

        # 4 图片数据 img, etag
        if image is not None and attrs['ltrb']:
            attrs['img'] = xlcv.get_sub(image, attrs['ltrb'])
            # attrs['etag'] = get_etag(attrs['img'])
            # TODO 这里目的其实就是让相同图片对比尽量相似，所以可以用dhash而不是etag

        # 5 中心点像素值 pixel
        p = attrs['center']
        if image is not None and p:
            attrs['pixel'] = tuple(image[p[1], p[0]].tolist()[::-1])

        # if 'rect' in attrs:
        #     del attrs['rect']  # 旧版的格式数据，删除

        return attrs

    def update_loclabel_img(self, loclabel, img, *, if_exists='update'):
        """ 添加一张图片数据

        :param if_exists:
            update，更新
            skip，跳过，不更新
        """
        loc, label = os.path.split(loclabel)
        h, w = img.shape[:2]

        # 1 如果loc图片不存在，则新建一个jpg图片
        update = True
        if loc not in self.data or not self.data[loc]:
            imfile = xlcv.write(img, File(loc, self.root, suffix='.jpg'))
            self.imfiles[loc] = imfile
            shape = LabelmeDict.gen_shape(label, [[0, 0], [w, h]])
            self.data[loc][label] = self.parse_shape(shape)
        # 2 不存在的标签，则在最后一行新建一张图
        elif label not in self.data[loc]:
            image = xlcv.read(self.imfiles[loc])
            height, width = image.shape[:2]
            assert width == w  # 先按行拼接，以后有时间可以扩展更灵活的拼接操作
            # 拼接，并重新存储为图片
            image = np.concatenate([image, img])
            xlcv.write(image, self.imfiles[loc])
            shape = LabelmeDict.gen_shape(label, [[0, height], [width, height + h]])
            self.data[loc][label] = self.parse_shape(shape)
        # 3 已有的图，则进行替换
        elif if_exists == 'update':
            image = xlcv.read(self.imfiles[loc])
            [x1, y1, x2, y2] = self.data[loc][label]['ltrb']
            image[y1:y2, x1:x2] = img
            xlcv.write(image, self.imfiles[loc])
        else:
            update = False

        if update:  # 需要实时保存到文件中
            self.write(loc)

    def write(self, loc):
        f = File(loc, self.root, suffix='.json')
        imfile = self.imfiles[loc]
        lmdict = LabelmeDict.gen_data(imfile)
        for label, ann in self.data[loc].items():
            a = ann.copy()
            DictTool.isub(a, ['img'])
            shape = LabelmeDict.gen_shape(json.dumps(a, ensure_ascii=False),
                                          a['points'], a['shape_type'],
                                          group_id=a['group_id'], flags=a['flags'])
            lmdict['shapes'].append(shape)
        f.write(lmdict, indent=2)

    def writes(self):
        for loc in self.data.keys():
            self.write(loc)

    def __getitem__(self, loclabel):
        """ 支持 self['主菜单/道友'] 的格式获取 self.data['主菜单']['道友'] 数据
        """
        # 注意功能效果：os.path.split('a/b/c') -> ('a/b', 'c')
        loc, label = os.path.split(loclabel)
        try:
            return self.data[loc][label]
        except KeyError:
            return None

    def __setitem__(self, loclabel, value):
        loc, label = os.path.split(loclabel)
        self.data[loc][label] = value


class NamedLocate(AutoGuiLabelData):
    """ 对有命名的标注数据进行定位

    注意labelme在每张图片上写的label值最好不要有重复，否则按字典存储后可能会被覆盖

    特殊取值：
    '@IMAGE_ID' 该状态图的标志区域，即游戏中在该位置出现此子图，则认为进入了对应的图像状态

    截图：screenshot，update_shot
    检查固定位像素：point2pixel，pixel_distance，check_pixel
    检查固定位图片：rect2img，img_distance，check_img
    检索图片：img2rects, img2rect，img2point，img2img
    操作：click，move_to
    高级：wait，check_click，wait_click
    """

    def __init__(self, root, *, region=None, grayscale=None, confidence=0.95, tolerance=20):
        """

        :param root: 标注数据所在文件夹
            有原图jpg、png和对应的json标注数据
        :param region: 可以限定窗口位置 (xywh格式)
        :param grayscale: 转成灰度图比较
        :param confidence: 用于图片比较，相似度阈值
        :param tolerance: 像素比较时能容忍的误差变化范围

        TODO random_click: 增加该可选参数，返回point时，进行上下左右随机扰动
        """
        super().__init__(root)
        self.grayscale = grayscale
        self.confidence = confidence
        self.tolerance = tolerance
        self.region = region
        self.last_shot = self.update_shot()

    def point_local2global(self, p):
        """ 一个self.region中相对坐标点p，转换到全局坐标位置 """
        if self.region:
            x, y = p
            x += self.region[0]
            y += self.region[1]
            return [x, y]
        else:
            return p

    def point_global2local(self, p):
        """ 一个self.region中相对坐标点p，转换到全局坐标位置 """
        if self.region:
            x, y = p
            x -= self.region[0]
            y -= self.region[1]
            return [x, y]
        else:
            return p

    def screenshot(self, region=None):
        """
        :param region: ltrb
        """
        if isinstance(region, str):
            region = ltrb2xywh(self[region]['ltrb'])
        im = pyautogui.screenshot(region=region)
        return xlpil.to_cv2_image(im)

    def update_shot(self, region=None):
        region = region or self.region
        self.last_shot = self.screenshot(region)
        return self.last_shot

    def point2pixel(self, point, haystack=None):
        """

        :param point: 支持输入 (x, y) 坐标，或者 loclabel
            后面很多参数都是类似，除了标准数据类型，同时支持 loclabel 定位模式

         官方的pixel版本，在windows有时候会出bug
         OSError: windll.user32.ReleaseDC failed : return 0
        """
        if isinstance(point, str):
            point = self[point]['center']
        if haystack is None:
            haystack = self.update_shot()
        return tuple(haystack[point[1], point[0]].tolist()[::-1])

    def rect2img(self, ltrb, haystack=None):
        """
        :param ltrb: 可以输入坐标定位，也可以输入 loclabel 定位
        """
        if isinstance(ltrb, str):
            ltrb = self[ltrb]['ltrb']
        l, t, r, b = ltrb
        if haystack is None:
            haystack = self.update_shot()
        return haystack[t:b, l:r]

    @classmethod
    def pixel_distance(cls, pixel1, pixel2):
        """ 返回最大的像素值差，如果相同返回0 """
        return max([abs(x - y) for x, y in zip(pixel1, pixel2)])

    @classmethod
    def img_distance(cls, img1, img2, *, tolerance=10):
        """ 返回距离：不相同像素占的百分比，相同返回0

        :param tolerance: 每个像素允许10的差距
        """
        cmp = np.array(abs(np.array(img1, dtype=int) - img2) > tolerance)
        return cmp.sum() / cmp.size

    def check_pixel(self, loclabel, haystack=None, *, tolerance=None):
        """ 判断对应位置上的像素值是否相同
        """
        tolerance = first_nonnone([tolerance, self.tolerance, 20])
        p1 = self[loclabel]['pixel']
        p2 = self.point2pixel(loclabel, haystack)
        return self.pixel_distance(p1, p2) < tolerance

    def check_img(self, loclabel, needle=None, *, grayscale=None, confidence=None):
        grayscale = first_nonnone([grayscale, self.grayscale, False])
        confidence = first_nonnone([confidence, self.confidence, 0.95])
        if needle is None:
            needle = self[loclabel]['img']
        elif isinstance(needle, str):
            needle = self[needle]['img']
        haystack = self.screenshot(loclabel)

        if confidence >= 1:
            return np.array_equal(needle, haystack)
        else:
            boxes = pyautogui.locateAll(needle, self.screenshot(loclabel),
                                        grayscale=grayscale, confidence=confidence)
            boxes = [xywh2ltrb(box) for box in list(boxes)]
            return len(boxes)

    def img2rects(self, img, haystack=None, *, grayscale=None, confidence=None):
        """ 根据预存的img数据，匹配出多个内容对应的所在的rect位置 """
        # 1 配置参数
        if isinstance(img, str):
            img = self[img]['img']
        grayscale = first_nonnone([grayscale, self.grayscale, False])
        confidence = first_nonnone([confidence, self.confidence, 0.95])
        # 2 查找子图
        if haystack is None:
            self.update_shot()
            haystack = self.last_shot
        boxes = pyautogui.locateAll(img, haystack,
                                    grayscale=grayscale,
                                    confidence=confidence)
        # 3 过滤掉重叠超过一半面积的框
        rects = ComputeIou.nms_ltrb([xywh2ltrb(box) for box in list(boxes)])
        return rects

    def img2rect(self, img, haystack=None, *, grayscale=None, confidence=None):
        """ img2rects的简化，只返回一个匹配框
        """
        rects = self.img2rects(img, haystack, grayscale=grayscale, confidence=confidence)
        return rects[0] if rects else None

    def img2point(self, img, haystack=None, *, grayscale=None, confidence=None):
        """ 将 img2rect 的结果统一转成点 """
        res = self.img2rect(img, haystack, grayscale=grayscale, confidence=confidence)
        if res:
            return np.array(np.array(res).reshape(2, 2).mean(axis=0), dtype=int).tolist()

    def img2img(self, img, haystack=None, *, grayscale=None, confidence=None):
        """ 找到rect后，返回匹配的目标img图片内容 """
        ltrb = self.img2rect(img, haystack, grayscale=grayscale, confidence=confidence)
        if ltrb:
            l, t, r, b = ltrb
            if haystack is not None:
                haystack = self.last_shot
            return haystack[t:b, l:r]

    def click(self, point, *, random_bias=0, back=False, wait_change=False, wait_seconds=0):
        """
        :param back: 点击后鼠标移回原坐标位置
        :param wait_change: 点击后，等待画面产生变化，才结束函数
            point为 loclabel 才能用这个功能
            注意这个只是比较loclabel的局部区域图片变化，不是全图的变化
        :param int random_bias: 允许在一定范围内随机点击点
        """
        # 1 判断类型
        if isinstance(point, str):
            loclabel = point
            point = self[loclabel]['center']
        else:
            loclabel = ''

        # 随机偏移点
        if random_bias:
            x, y = point
            x += random.randint(-random_bias, +random_bias)
            y += random.randint(-random_bias, +random_bias)
            point = (x, y)

        # 2 鼠标移动
        pos0 = pyautogui.position()
        pyautogui.click(*self.point_local2global(point))
        if back:
            pyautogui.moveTo(*pos0)  # 恢复鼠标原位置

        # 3 等待画面变化
        if loclabel and wait_change:
            func = lambda: self.img_distance(self[loclabel]['img'],
                                             self.rect2img(loclabel)) > self.confidence
            xlwait(func)
        if wait_seconds:
            time.sleep(wait_seconds)

        return point

    def move_to(self, point):
        if isinstance(point, str):
            point = self[point]['center']
        pyautogui.moveTo(*self.point_local2global(point))

    def drag_to(self, loclabel, direction=0, duration=1, *, to_tail=False, **kwargs):
        """ 在loclabel区域内进行拖拽操作

        :param direction: 方向，0123分别表示向"上/右/下/左"拖拽。
        :param to_tail:
            False，默认只拖拽一次
            True，一直拖拽到末尾，即内容不再变化为止
        """

        def core():
            x = self[loclabel]
            l, t, r, b = x['ltrb']
            x2, y2 = x['center']

            self.move_to(loclabel)
            if direction == 0:
                y2 = (y2 + t) // 2
            elif direction == 1:
                x2 = (x2 + r) // 2
            elif direction == 2:
                y2 = (y2 + b) // 2
            elif direction == 3:
                x2 = (x2 + l) // 2
            else:
                raise ValueError

            x2, y2 = self.point_local2global([x2, y2])
            pyautogui.dragTo(x2, y2, duration=duration, **kwargs)

        if to_tail:
            last_im = self.rect2img(loclabel)
            while True:
                core()
                time.sleep(1)
                new_im = self.rect2img(loclabel)
                if self.img_distance(last_im, new_im) > 0.95:  # 95%的像素相似
                    break
                last_im = new_im
        else:
            core()

    def wait(self, loclabel, *, fixpos=False, limit=None, interval=1):
        """ 封装的比较高层的功能 """
        # dprint(loclabel, fixpos)
        if fixpos:
            xlwait(lambda: self.check_img(loclabel))
            point = self[loclabel]['center']
        else:
            point = xlwait(lambda: self.img2point(loclabel), limit=limit, interval=interval)
        return point

    def wait_leave(self, loclabel, *, fixpos=False, limit=None, interval=1):
        """ 封装的比较高层的功能 """
        if fixpos:
            xlwait(lambda: self.check_img(loclabel))
            point = self[loclabel]['center']
        else:
            point = xlwait(lambda: not self.img2point(loclabel), limit=limit, interval=interval)
        return point

    def check_click(self, loclabel, *, back=False, wait_change=False):
        point = self.img2point(loclabel)
        if point:
            self.click(point, back=back, wait_change=wait_change)

    def wait_click(self, loclabel, *, fixpos=False, back=False, wait_change=False):
        """ 封装的比较高层的功能 """
        point = self.wait(loclabel, fixpos=fixpos)
        self.click(point, back=back, wait_change=wait_change)

    def point2loclabels(self, point, loclabels):
        """ 判断loc这张图里有哪些标签覆盖到了point这个点

        :param loclabels: 可以输入一张图的定位
            也可以输入多个loclabel清单，会返回所有满足的loclabel名称
        """
        from shapely.geometry import Point

        # 1 loclabels
        if not is_list_like(loclabels):
            loclabels = [loclabels]
        point = Point(*point)

        # 2 result
        res = []
        for loclabel in loclabels:
            if loclabel in self.data:
                for k, v in self.data.items():
                    if 'ltrb' in v:
                        if point.within(ShapelyPolygon.gen(v['ltrb'])):
                            res.append(f'{loclabel}/{k}')
            else:
                v = self[loclabel]
                if 'ltrb' in v:
                    if point.within(ShapelyPolygon.gen(v['ltrb'])):
                        res.append(f'{loclabel}')

        return res

    def point2loclabel(self, point, loclabels):
        """ 只返回第一个匹配结果
        """
        from shapely.geometry import Point

        # 1 loclabels
        if not is_list_like(loclabels):
            loclabels = [loclabels]
        point = Point(*point)

        # 2 result
        for loclabel in loclabels:
            if loclabel in self.data:
                for k, v in self.data.items():
                    if 'ltrb' in v:
                        if point.within(ShapelyPolygon.gen(v['ltrb'])):
                            return f'{loclabel}/{k}'
            else:
                v = self[loclabel]
                if 'ltrb' in v:
                    if point.within(ShapelyPolygon.gen(v['ltrb'])):
                        return f'{loclabel}'


class PosTran:
    """ 坐标位置变换

    应用场景： 原来在A窗口下的点p和区域r，
        在窗口位置、大小改成B后，p和r的坐标
    """

    def __init__(self, w1, w2):
        """
        :param w1: 原窗口位置 (x, y, w, h)
        :param w2: 新窗口位置
        """
        self.w1 = w1
        self.w2 = w2

    @classmethod
    def point2point(cls, w1, p1, w2):
        """ 窗口w1切到w2，原点p1变换到坐标p2
        """
        x1, y1, w1, h1 = w1
        x2, y2, w2, h2 = w2
        x = x2 + (p1[0] - x1) * w2 / w1
        y = y2 + (p1[1] - y1) * h2 / h1
        return round(x), round(y)

    def w1point(self, w2point):
        return self.point2point(self.w2, w2point, self.w1)

    def w2point(self, w1point):
        """旧窗口中的点坐标转回新窗口点坐标"""
        return self.point2point(self.w1, w1point, self.w2)

    @classmethod
    def dx2dx(cls, w1, dx1, w2):
        """ 宽度偏移量的变化
        """
        return round(dx1 * w2[2] / w1[2])

    def w1dx(self, dx2):
        return self.dx2dx(self.w2, dx2, self.w1)

    def w2dx(self, dx1):
        return self.dx2dx(self.w1, dx1, self.w2)

    @classmethod
    def dy2dy(cls, w1, dy1, w2):
        return round(dy1 * w2[3] / w1[3])

    def w1dy(self, dy2):
        return self.dy2dy(self.w2, dy2, self.w1)

    def w2dy(self, dy1):
        return self.dy2dy(self.w1, dy1, self.w2)

    @classmethod
    def region2region(cls, w1, r1, w2):
        """ 窗口w1切到w2，原区域r1变换到坐标r2
        """
        x, y = cls.point2point(w1, r1[:2], w2)
        w = round(r1[2] * w2[2] / w1[2])
        h = round(r1[3] * w2[3] / w1[3])
        return x, y, w, h

    def w1region(self, w2region):
        return self.region2region(self.w2, w2region, self.w1)

    def w2region(self, w1region):
        return self.region2region(self.w1, w1region, self.w2)


def lookup_mouse_position():
    """ 查看鼠标位置的工具
    """
    from keyboard import read_key

    left_top_point = None
    while True:
        k = read_key()
        if k == 'ctrl':
            # 定位当前鼠标坐标位置
            # 也可以用来定位区域时，先确定区域的左上角点
            left_top_point = pyautogui.position()
            print('坐标：', *left_top_point)
        elif k == 'alt':
            # 定位区域的右下角点，并输出区域
            p = pyautogui.position()
            print('区域(x y w h)：', left_top_point.x, left_top_point.y, p.x - left_top_point.x, p.y - left_top_point.y)
        elif k == 'esc':
            break
        # keyboard的监控太快了，需要暂停一下
        time.sleep(0.4)


def lookup_mouse_position2(w1, w2, reverse=False):
    """ 涉及到窗口位置、大小调整时的坐标计算
        当前在窗口w1的坐标，切换到w2时的坐标
    :param reverse: 当前窗口实际是变换后的w2，输出的时候需要按照 w1 -> w2 的格式展示
        代码实现的时候，其实只要把输出的内容对调即可

    该函数是lookup_mouse_postion的进阶版，两个函数还是先不做合并，便于理解维护
    """
    import keyboard
    postran = PosTran(w1, w2)

    left_top_point = None
    while True:
        k = keyboard.read_key()
        if k == 'ctrl':
            # 定位当前鼠标坐标位置
            # 也可以用来定位区域时，先确定区域的左上角点
            left_top_point = pyautogui.position()
            p1, p2 = left_top_point, postran.w2point(left_top_point)
            if reverse: p1, p2 = p2, p1
            print('坐标：', *p1, '-->', *p2)
        elif k == 'alt':
            # 定位区域的右下角点，并输出区域
            p = pyautogui.position()
            r1 = [left_top_point.x, left_top_point.y, p.x - left_top_point.x, p.y - left_top_point.y]
            r2 = postran.w2region(r1)
            if reverse: r1, r2 = r2, r1
            print('区域(x y w h)：', *r1, '-->', *r2)
        elif k == 'esc':
            break
        time.sleep(0.4)


def type_text(text):
    """ 打印出文本内容

    相比pyautogui.write，这里支持中文等unicode格式

    这种需求一般也可以用剪切板实现，但是剪切板不够独立轻量，可能会有意想不到的副作用
    """
    check_install_package('pynput')
    from pynput.keyboard import Controller

    keyboard = Controller()
    keyboard.type(text)


def clipboard_decorator(rtype='text', *, copy=True, paste=False, typing=False):
    """ 装饰器，能方便地将函数的返回值转移到剪切板文本、富文本，甚至typing打印出来。

    Args:
        rtype: 函数返回值类型
        paste: 是否执行剪切板的粘贴功能
        typing: 是否键入文本内容

    Returns:

    """
    from bs4 import BeautifulSoup
    import pyautogui
    import klembord

    def decorator(func):
        def wrapper(*args, **kwargs):
            # 1 运行函数获得结果
            s = func(*args, **kwargs)

            # 2 复制到剪切板
            if copy:
                if rtype == 'text' and isinstance(s, str):
                    klembord.set_text(s)
                elif rtype == 'html' and isinstance(s, str):
                    s0 = BeautifulSoup(s, 'lxml').text
                    klembord.set_with_rich_text(s0, s)
                elif rtype == 'dict' and isinstance(s, dict):
                    klembord.set(s)

            # 3 输出
            if paste:
                pyautogui.hotkey('ctrl', 'v')  # 目前来看就这个方法最靠谱
            if typing:
                type_text(s)

            return s

        return wrapper

    return decorator


def get_clipboard_content(rich=False, *, head=False):
    """ klembord的get_with_rich_text获取富文本有点小问题，所以自己重写了一个版本
    好在有底层接口，所以也不难改

    :param rich: 是否返回html格式的富文本，默认True
    :param head: 富文本时，是否要返回<html><body>的标记
    """
    import re
    import html

    import klembord
    from klembord import W_HTML, W_UNICODE

    if not rich:
        return klembord.get_text()

    content = klembord.get((W_HTML, W_UNICODE))
    html_content = content[W_HTML]
    if html_content:
        html_content = html_content.decode('utf8')
        html_content = re.search(r'<body>\s*(?:<!--StartFragment-->)?(.+?)(?:<!--EndFragment-->)?\s*</body>',
                                 html_content).group(1)
    else:
        # 没有html的时候，可以封装成html格式返回
        text = content[W_UNICODE].decode('utf16')
        html_content = html.escape(text)

    if head:
        html_content = '<html><body>' + html_content + '</body></html>'

    return html_content


def set_clipboard_content(content, *, rich=False):
    """ 对klembord剪切板的功能做略微的简化 """
    from bs4 import BeautifulSoup
    import klembord

    if rich:
        klembord.set_with_rich_text(BeautifulSoup(content, 'lxml').text, content)
    else:
        klembord.set_text(content)


def grab_pixel_color(radius=0):
    """
    :param radius: 以中心像素为例，计算矩阵范围内的颜色分布情况（未实装）
    """
    from pyxllib.cv.rgbfmt import RgbFormatter
    last_pos = None
    print('精确全体/精确中文')

    while True:
        pos = pyautogui.position()
        if pos == last_pos:
            continue
        else:
            last_pos = pos

        color = RgbFormatter(*pyautogui.pixel(*pos))
        desc = color.relative_color_desc(color_range=2, precise_mode=True)
        print('\r' + f'{color.to_hex()} {desc}', end='', flush=True)
        time.sleep(0.1)


def grab_monitor(order=0, to_cv2=True):
    """ 截屏

    :param int order:
        默认0，截取全部屏幕
        1，截取第1个屏幕
        2，截取第2个屏幕
        ...
    :param to_cv2:
        True, 转成cv2格式的图片
        False, 使用pil格式的图片
    """
    from PIL import Image
    from pyxllib.xlcv import PilImg

    with mss.mss() as sct:
        assert order < len(sct.monitors), '屏幕下标越界'
        monitor = sct.monitors[order]
        sct_img = sct.grab(monitor)
        img = Image.frombytes("RGB", sct_img.size, sct_img.bgra, "raw", "BGRX")
        if to_cv2:
            img = PilImg.read(img).to_cv2_image()  # 返回是np矩阵
        return img


def list_windows(mode=1):
    """ 列出所有窗口

    :param mode:
        0, 列出所有窗口
        1, 只列出有名称，有宽、高的窗口
        str, 指定名称的窗口
            注意，当结果有多个的时候，一般后面那个才是紧贴边缘的窗口位置

    :return: [[窗口名, xywh], [窗口2, xywh]...]
    """
    import win32gui

    ls = []

    def callback(hwnd, extra):
        l, t, r, b = win32gui.GetWindowRect(hwnd)
        w, h = r - l, b - t
        name = win32gui.GetWindowText(hwnd)
        if mode == 0:
            ls.append([name, [l, t, w, h]])
        elif mode == 1 and name and w and h:
            ls.append([name, [l, t, w, h]])
        elif isinstance(mode, str) and name == mode:
            ls.append([name, [l, t, w, h]])

    win32gui.EnumWindows(callback, None)
    return ls


class UiTracePath:
    """ 在window窗口上用鼠标勾画路径 """

    @classmethod
    def from_polygon(cls, points, duration_per_circle=3, num_circles=1):
        """
        让鼠标顺时针围绕任意多边形移动。

        :param points: 多边形顶点的坐标列表，按顺时针或逆时针顺序排列
        :param duration_per_circle: 每圈耗时
        :param num_circles: 总圈数
        """
        num_points = len(points)
        segment_duration = duration_per_circle / num_points  # 每条边的耗时

        for _ in range(num_circles):
            # 顺时针移动鼠标
            for i in range(num_points):
                start = points[i]
                pyautogui.moveTo(start[0], start[1])
                end = points[(i + 1) % num_points]
                # 使用 moveTo 的 duration 参数控制每条边的移动时间
                pyautogui.moveTo(end[0], end[1], duration=segment_duration)

    @classmethod
    def from_ltrb(cls, ltrb, **kwargs):
        points = ltrb2polygon(ltrb)
        return cls.from_polygon(points, **kwargs)


if __name__ == '__main__':
    with TicToc(__name__):
        agld = AutoGuiLabelData(r'D:\slns\py4101\py4101\touhou\label')
        agld.writes()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Date   : 2024/07/31

import os
import re

import requests
import pandas as pd


class WpsOnlineBook:
    """ wps的"脚本令牌"调用模式

    官方文档：https://airsheet.wps.cn/docs/apitoken/api.html
    """

    def __1_基础功能(self):
        pass

    def __init__(self, file_id, script_id=None, *, token=None):
        self.headers = {
            'Content-Type': "application/json",
            'AirScript-Token': token or os.getenv('WPS_SCRIPT_TOKEN', ''),
        }
        self.file_id = file_id
        self.default_script_id = script_id

    def post_request(self, url, payload):
        """
        发送 POST 请求到指定的 URL 并返回响应结果
        """
        try:
            resp = requests.post(url, json=payload, headers=self.headers)
            resp.raise_for_status()  # 如果请求失败会抛出异常
            return resp.json()
        except requests.exceptions.RequestException as e:
            print(f"请求失败: {e}")
            return None

    def run_script(self, script_id=None, context_argv=None, sync=True):
        """ 原本执行 WPS 脚本并返回执行结果

        :param script_id: 脚本 ID
        :param context_argv: 脚本参数 (可选)
            context本来能支持这些参数的：
                dict argv: 传入的上下文参数对象，比如传入{name: 'xiaomeng', age: 18}， 在 AS 代码中可通过Context.argv.name获取到传入的值
                str sheet_name: et,ksheet 运行时所在表名
                str range: et,ksheet 运行时所在区域，例如$B$156
                str link_from: et,ksheet 点击超链接所在单元格
                str db_active_view: db 运行时所在 view 名
                str db_selection: db 运行时所在选区
            但是，sheet_name, range等，并看不到对运行代码有什么实质影响，不影响active，而且as里也引用不了sheet_name等值
                所以退化，简化为只要传入content_argv参数就行
        :param sync:
            True, 同步运行
            False, 异步运行

        这个接口跟普通as一样，运行有30秒时限
        """
        url = f"https://www.kdocs.cn/api/v3/ide/file/{self.file_id}/script/{script_id}/{'sync_task' if sync else 'task'}"
        payload = {
            "Context": {'argv': context_argv or {}}
        }
        res = self.post_request(url, payload)
        if sync:
            return res['data']['result']
        else:
            return res

    def __2_封装的更高级的接口(self):
        """ 这系列的功能需要配套这个框架范式使用：
        https://github.com/XLPRUtils/pyxllib/blob/master/pyxllib/text/airscript.js
        """
        pass

    def run_func(self, func_name, *args):
        """ 我自己常用的jsa框架，jsa那边已经简化了对接模式，所以一般都只用这个高级的接口即可
        （旧函数名run_script2不再使用）
        """
        return self.run_script(self.default_script_id, context_argv={'funcName': func_name, 'args': args})

    def get_sheet_data(self, sheet_name, fields, data_row=0, filter_empty_rows=True, *,
                       return_mode='pd') -> pd.DataFrame:
        """ 获取某张sheet表格数据

        :param sheet_name: sheet表名
        :param list[str] fields: 字段名列表
        :param int data_row: 数据起始行，详细用法见packTableDataFields
        :param return_mode: 'pd' or 'json'
        """
        data = self.run_func('packTableDataFields', sheet_name, fields, data_row, filter_empty_rows)
        if return_mode == 'json':
            return data
        elif return_mode == 'pd':
            return pd.DataFrame(data)

    def write_arr(self, rows, start_cell, batch_size=None):
        """ 把一个二维数组数据写入表格

        :param rows: 一个n*m的数据
        :param start_cell: 写入的起始位置，例如'A1'，也可以使用Sheet1!A1的格式表示具体的表格位置
        :param batch_size: 为了避免一次写入内容过多，超时写入失败，可以分成多批运行
            这里写每批的数据行数
            默认表示一次性全部提交
        :return:
        """
        if batch_size is None:
            batch_size = len(rows)  # 如果未指定批次大小，一次性写入所有行

        def func(m):
            return str(int(m.group()) + batch_size)

        current_cell = start_cell
        for start in range(0, len(rows), batch_size):
            end = start + batch_size
            batch_rows = rows[start:end]
            self.run_func('writeArrToSheet', batch_rows, current_cell)
            current_cell = re.sub(r'\d+$', func, current_cell)


if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/11/05

"""
以uiautomation为核心的相关工具库
"""

import os
import sys
import textwrap
import time
from typing import Iterable, Callable, List
import subprocess
import tempfile

import psutil
import pandas as pd
from fastcore.basics import GetAttr

from loguru import logger
# ui组件大多是树形组织结构，auto库自带树形操作太弱。没有专业的树形结构库，能搞个毛线。
from anytree import NodeMixin

import ctypes
from ctypes import wintypes

if sys.platform == 'win32':
    import win32con
    import win32gui
    import win32process
    import win32clipboard

    import uiautomation as uia
    from uiautomation import WindowControl


def __1_clipboard_utils():
    pass


def retry_on_failure(max_retries: int = 5):
    """
    一个装饰器，用于在失败时重试执行被装饰的函数。

    Args:
        max_retries (int): 最大重试次数。

    Returns:
        Callable: 包装后的函数。
    """

    def decorator(func: Callable):
        def wrapper(*args, **kwargs):
            for attempt in range(max_retries):
                try:
                    if func(*args, **kwargs):
                        return True
                except Exception as e:
                    time.sleep(.05)
                    print(f"Attempt {attempt + 1} failed: {e}")
            return False

        return wrapper

    return decorator


def set_clipboard_data(fmt: int, buf: ctypes.Array) -> bool:
    """
    将数据设置到Windows剪切板中。

    Args:
        fmt (int): 数据格式，例如 win32clipboard.CF_HDROP。
        buf (ctypes.Array): 要设置到剪切板的数据。

    Returns:
        bool: 操作成功返回 True，否则返回 False。
    """
    try:
        win32clipboard.OpenClipboard()
        win32clipboard.EmptyClipboard()
        win32clipboard.SetClipboardData(fmt, buf)
        return True
    except Exception as e:
        print(f"Error setting clipboard data: {e}")
        return False
    finally:
        win32clipboard.CloseClipboard()


def get_clipboard_files() -> List[str]:
    """
    获取剪切板中的文件路径列表。

    Returns:
        List[str]: 包含剪切板中文件路径的列表，如果没有文件路径或操作失败，返回空列表。
    """
    try:
        win32clipboard.OpenClipboard()
        if win32clipboard.IsClipboardFormatAvailable(win32clipboard.CF_HDROP):
            return list(win32clipboard.GetClipboardData(win32clipboard.CF_HDROP))
        else:
            return list()
    finally:
        win32clipboard.CloseClipboard()


@retry_on_failure(max_retries=5)
def validate_clipboard_files(file_paths: Iterable[str], fmt: int, buf: ctypes.Array) -> bool:
    """
    验证剪切板中的文件路径是否与给定的文件路径一致。

    Args:
        file_paths (Iterable): 一个包含文件路径的可迭代对象，每个路径都是一个字符串。
        fmt (int): 数据格式，例如 win32clipboard.CF_HDROP。
        buf (ctypes.Array): 要验证的剪切板数据。

    Returns:
        bool: 如果剪切板中的文件路径与给定的文件路径一致，则返回 True

    Raises:
        ValueError: 如果剪切板文件路径与给定文件路径不一致。
    """
    # 设置文件到剪切板
    set_clipboard_data(fmt, buf)
    # 验证剪切板中的文件路径是否与给定的文件路径一致
    if set(get_clipboard_files()) == set(file_paths):
        return True
    raise ValueError("剪切板文件路径不对哇！")


def copy_files_to_clipboard(file_paths: Iterable[str]) -> bool:
    """
    将一系列文件路径复制到Windows剪切板。这允许用户在其他应用程序中，如文件资源管理器中粘贴这些文件。

    Args:
        file_paths (Iterable): 一个包含文件路径的可迭代对象，每个路径都是一个字符串。

    Returns:
        bool: 如果成功将文件路径复制到剪切板，则返回 True，否则返回 False
    """
    # 定义所需的 Windows 结构和函数
    CF_HDROP = 15

    class DROPFILES(ctypes.Structure):
        _fields_ = [("pFiles", wintypes.DWORD),
                    ("pt", wintypes.POINT),
                    ("fNC", wintypes.BOOL),
                    ("fWide", wintypes.BOOL)]

    offset = ctypes.sizeof(DROPFILES)
    length = sum(len(p) + 1 for p in file_paths) + 1
    size = offset + length * ctypes.sizeof(wintypes.WCHAR)
    buf = (ctypes.c_char * size)()
    df = DROPFILES.from_buffer(buf)
    df.pFiles, df.fWide = offset, True
    for path in file_paths:
        path = os.path.normpath(path)
        array_t = ctypes.c_wchar * (len(path) + 1)
        path_buf = array_t.from_buffer(buf, offset)
        path_buf.value = path
        offset += ctypes.sizeof(path_buf)
    buf[offset:offset + ctypes.sizeof(wintypes.WCHAR)] = b'\0\0'

    # 验证文件是否成功复制到剪切板
    return validate_clipboard_files([os.path.normpath(file) for file in file_paths], CF_HDROP, buf=buf)


def __2_窗口功能():
    pass


def get_windows_info():
    """ 得到当前机器的全部窗口信息清单 """
    window_list = []

    def get_all_hwnd(hwnd, mouse):
        thread_id, process_id = win32process.GetWindowThreadProcessId(hwnd)
        proc = psutil.Process(process_id)

        is_window = win32gui.IsWindow(hwnd)
        is_enabled = win32gui.IsWindowEnabled(hwnd)
        is_visible = win32gui.IsWindowVisible(hwnd)
        text = win32gui.GetWindowText(hwnd)

        data = {
            'proc_name': proc.name(),
            'process_id': process_id,
            'thread_id': thread_id,
            'hwnd': hwnd,
            'ClassName': win32gui.GetClassName(hwnd),
            'ControlTypeName': '',
            'WindowText': text,
            'IsWindow': is_window,
            'IsWindowEnabled': is_enabled,
            'IsWindowVisible': is_visible
        }

        if not data['proc_name'].endswith('.tmp') and is_visible:
            ctrl = uia.ControlFromHandle(hwnd)
            data['ControlTypeName'] = ctrl.ControlTypeName

        window_list.append(data)

    win32gui.EnumWindows(get_all_hwnd, 0)
    return pd.DataFrame(window_list)


def find_ctrl(class_name=None, name=None, **kwargs):
    if class_name is not None:
        kwargs['ClassName'] = class_name
    if name is not None:
        kwargs['Name'] = name
    ctrl = uia.WindowControl(**kwargs)
    return ctrl


class UiCtrlNode(NodeMixin, GetAttr, WindowControl):
    _default = 'ctrl'

    def __0_构建(self):
        pass

    def __init__(self, ctrl, parent=None, *, build_depth=-1):
        """
        :param ctrl: 当前节点
        :param parent: 父结点
        :param build_depth: 自动构建多少层树节点，默认-1表示构建全部节点
        """
        # 初始化节点信息
        self.ctrl = ctrl
        # 试过了，没用，因为新找出来的都是新构建的类，找不到proxy的
        # self.ctrl.proxy: 'UiCtrlNode' = self  # 再给其扩展一个proxy属性，指向其升级过的对象
        self.ctrl_type = ctrl.ControlTypeName
        self.text = ctrl.Name
        self.parent = parent  # 指定父节点，用于形成树结构

        # 自动递归创建子节点
        self.build_children(build_depth)

    @classmethod
    def init_from_name(cls, class_name=None, name=None, *, build_depth=-1, **kwargs):
        ctrl = find_ctrl(class_name=class_name, name=name, **kwargs)
        return cls(ctrl, build_depth=build_depth)

    def activate(self, check_seconds=2):
        """ 激活当前窗口
        """
        while True:
            # todo 这种限定情况并不严谨，有概率出现重复的~
            hwnd = win32gui.FindWindow(self.ctrl.ClassName, self.text)
            # logger.info(hwnd)
            if not hwnd:
                return

            if win32gui.GetForegroundWindow() != hwnd:
                win32gui.ShowWindow(hwnd, win32con.SW_RESTORE)

                try:
                    # 在这里执行SetForegroundWindow，只有程序的第1次运行有效，之后就会被很多全屏类的应用占用最前置，覆盖不了了
                    # 为了解决这问题，就只能暴力每次都新开一个程序来执行这个SetForegroundWindow操作
                    subprocess.run(
                        [sys.executable, "-c", f"import win32gui; win32gui.SetForegroundWindow({hwnd})"],
                        stdout=subprocess.PIPE,
                    )
                except Exception as e:
                    pass
                # 理论上并不需要等待，但加个等待，有助于稳定性检测，如果当前窗口在check_seconds秒内频繁切换，
                #   使用activate虽然激活了，但并不安全，只有check_seconds秒内维持稳定在这个窗口，再进行下游任务会更好
                time.sleep(check_seconds)
            else:
                break

    def build_children(self, build_depth=-1, child_node_class=None):
        """ 创建并添加子节点到树中 """
        if build_depth == 0:
            return
        self.children = []  # 删除现有的所有子结点
        child_node_class = child_node_class or self.__class__
        for child_ctrl in self.ctrl.GetChildren():
            child_node_class(child_ctrl, parent=self, build_depth=build_depth - 1)

    def __1_调试(self):
        pass

    def trace_rect(self, duration_per_circle=2, num_circles=1):
        """ 用鼠标勾画出当前组件的矩形位置区域 """
        from pyxllib.autogui.autogui import UiTracePath

        rect = self.ctrl.BoundingRectangle
        ltrb = [rect.left, rect.top, rect.right, rect.bottom]
        UiTracePath.from_ltrb(ltrb,
                              duration_per_circle=duration_per_circle,
                              num_circles=num_circles)

    def __2_功能(self):
        pass

    def __getattr__(self, item):
        # 尝试从self.ctrl中获取属性
        return getattr(self.ctrl, item)

    def __getitem__(self, index):
        """ 通过索引直接访问子节点

        ui操作经常要各种结构化的访问，加个这个简化引用方式
        """
        try:
            return self.children[index]
        except IndexError:  # 如果出现下标错误，需要自动重新刷新所有控件
            # self.parent重建是不够的，但我也不知道为什么self.root重建后就可以了
            # 我的理解是重建后self自己不是都不存在了？
            self.root.build_children()
            # 应该在有些情况下self.root重建还能继续使用，但有些特殊情况应该会炸
            return self.children[index]

    def get_ctrl_hash_tag(self, level=1):
        """ 生成节点的哈希字符串，以反映子树结构，一般用来对节点做分类及映射到对应处理函数 """
        # 当前节点的类型标识符
        hash_strs = [f"{level}{self.ctrl_type[0].lower()}"]
        # 遍历所有子节点，递归生成子节点的哈希值
        for child in self.children:
            hash_strs.append(f"{child.get_ctrl_hash_tag(level + 1)}")
        return ''.join(hash_strs)

    def __3_展示(self):
        pass

    def _format_text(self, text):
        """ 将换行替换为空格的小工具方法 """
        return text.replace('\n', ' ')

    def __repr__(self):
        """ 用于在打印节点时显示关键信息 """
        return f"UiNode(ctrl_type={self.ctrl_type}, text={self._format_text(self.text)})"

    def render_tree(self):
        """ 展示以self为根节点的整体内容结构 """
        # 1 渲染自身
        line = [self.ctrl_type]
        line.append(self._format_text(self.text))

        # 加上控件的坐标信息
        rect = self.ctrl.BoundingRectangle
        line.append(f"[{rect.left}, {rect.top}, {rect.right}, {rect.bottom}]")

        # 我的hash值
        tag = self.get_ctrl_hash_tag()
        if len(tag) <= 64:
            line.append(tag)

        lines = [' '.join(line)]

        # 2 子结点情况
        for child in self.children:
            line = child.render_tree()
            line = textwrap.indent(line, '    ')
            lines.append(line)
        return '\n'.join(lines)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/10/31

"""
代码参考(抄自)：https://github.com/Frica01/WeChatMassTool
"""

import sys
import time
from copy import deepcopy
from typing import (Iterable, Callable, List)
from typing import Optional
from unittest.mock import MagicMock

from loguru import logger

if sys.platform == 'win32':
    import uiautomation as auto
    import win32con
    import win32gui

from pyxllib.prog.filelock import get_autoui_lock

from pyxllib.autogui.uiautolib import find_ctrl, UiCtrlNode, copy_files_to_clipboard


def __1_config():
    pass


class WeChatConfig:
    WeChat_PROCESS_NAME = 'WeChat.exe'
    APP_NAME = 'WeChatMassTool'
    APP_PROCESS_NAME = 'WeChatMassTool.exe'
    APP_LOCK_NAME = 'WeChatMassTool.lock'
    WINDOW_NAME = '微信'
    WINDOW_CLASSNAME = 'WeChatMainWndForPC'


# WeChat = WeChatConfig


class IntervalConfig:
    BASE_INTERVAL = 0.1  # 基础间隔（秒）
    SEND_TEXT_INTERVAL = 0.05  # 发送文本间隔（秒）
    SEND_FILE_INTERVAL = 0.25  # 发送文件间隔（秒）
    MAX_SEARCH_SECOND = 0.1
    MAX_SEARCH_INTERVAL = 0.05


Interval = IntervalConfig


def __2_window_utils():
    pass


def minimize_wechat(class_name, name):
    """
    关闭Windows窗口

    Args:
        name(str):  进程名
        class_name(str):  进程class_name

    Returns:

    """
    hwnd = win32gui.FindWindow(class_name, name)
    if win32gui.IsWindowVisible(hwnd):
        win32gui.SendMessage(hwnd, win32con.WM_CLOSE, 0, 0)


def wake_up_window(class_name, name):
    """
    唤醒Windows窗口

    Args:
        name(str):  进程名
        class_name(str):  进程class_name

    Returns:

    """
    if hwnd := win32gui.FindWindow(class_name, name):
        # 恢复窗口
        win32gui.ShowWindow(hwnd, win32con.SW_RESTORE)
        # 检查窗口是否已在前
        if win32gui.GetForegroundWindow() != hwnd:
            # 尝试将窗口置前
            try:
                win32gui.SetForegroundWindow(hwnd)
            except Exception as e:
                pass
                # print(f"尝试将窗口置前时出错: {e}")


def is_window_visible(class_name, name):
    """
    唤醒Windows 窗口的可见性

    Args:
        name(str):  进程名
        class_name(str):  进程class_name

    Returns:

    """
    if hwnd := win32gui.FindWindow(class_name, name):
        # 判断窗口可见性
        if win32gui.IsWindowVisible(hwnd):
            return True
    return False


def __3_wechat():
    pass


class MsgNode(UiCtrlNode):
    """ 一条消息的节点 """

    def __init__(self, msg_ctrl, parent=None, *, build_depth=-1):
        super().__init__(msg_ctrl, parent, build_depth=build_depth)
        self.time = None  # 每条消息都建立一个最近出现的时间作为时间戳？
        self.msg_type = None  # 消息类型：system, receive, send
        self.content_type = None  # 内容类型：text、image、file、time

        self.user = None  # 用户名
        self.user2 = None  # 群聊时用户会多一个群聊昵称
        self.cite_text = None  # 引用的内容，引用的不一定是文本，但是统一转成文本显示的

        self.render_text = None  # 供人预览，和简化给ai查看的格式

    def build_children(self, build_depth, child_node_class=None):
        if build_depth == 0:
            return
        self.children = []
        child_node_class = child_node_class or UiCtrlNode
        for child_ctrl in self.ctrl.GetChildren():
            child_node_class(child_ctrl, parent=self, build_depth=build_depth - 1)

    def init(self):
        """ 解析出更精细的结构化信息 """
        from pyxllib.autogui.wechat_msg import msg_parsers

        tag = self.get_ctrl_hash_tag()
        if tag in msg_parsers:
            msg_parsers[tag](self)

    def render_tree(self):
        """ 展示以self为根节点的整体内容结构 """
        # 1 未解析过的节点，原样输出树形结构供添加解析
        if not self.msg_type:
            return super().render_tree()

        # 2 已设置好预览文本的直接输出
        if self.render_text:
            return self.render_text

        # 3 否则用一套通用的机制渲染输出
        fmt = self._format_text
        # 定义类型到符号的映射
        type_to_symbol = {
            'send': '↑',
            'receive': '↓',
            'system': '⚙️'  # 假设 'system' 类型对应的符号是 ⚙️
        }
        # 根据 msg_type 获取相应的符号
        msg = [f"{type_to_symbol.get(self.msg_type, self.msg_type)}"]

        if self.user:
            msg.append(f"{self.user}: ")
        msg.append(fmt(self.text))

        if self.cite_text:
            msg.append(f"【引用】{fmt(self.cite_text)}")
        return ' '.join(msg)

    def is_match(self, msg_node):
        """ 两条msg_node是否对应的上 """
        from datetime import datetime

        # 1 比如内容一致
        if self.render_tree() == msg_node.render_tree():
            return 2  # 强一致

        # 2 或者"撤回"格式对应的上
        is_recall = (self.content_type == 'recall' or msg_node.content_type == 'recall')
        user_matches = is_recall and (self.user == msg_node.user or self.user2 == msg_node.user2)
        if is_recall and user_matches:
            return 1  # 弱一致

        # 3 判断是否为 'system'、'time' 类型
        is_system_time_type = (self.msg_type == msg_node.msg_type == 'time')
        if is_system_time_type:
            # 检查时间匹配，确保都是 datetime 类型，并只比较年月日时分部分
            same_time = (
                    isinstance(self.time, datetime) and
                    isinstance(msg_node.time, datetime) and
                    self.time.strftime('%Y-%m-%d %H:%M') == msg_node.time.strftime('%Y-%m-%d %H:%M')
            )
            if same_time:
                return 2

            if self.text == '以下是新消息' or msg_node.text == '以下是新消息':
                return 1

        # 4 如果有一条.content_type=='button_more'，也直接弱匹配
        if self.content_type == 'button_more' or msg_node.content_type == 'button_more':
            return 1


class MsgBoxNode(UiCtrlNode):
    """ 当前会话消息窗的节点 """

    def __init__(self, chat_box_ctrl, parent=None, *, build_depth=-1):
        super().__init__(chat_box_ctrl, parent, build_depth=build_depth)

    def build_children(self, build_depth=-1, child_node_class=None):
        # 1 遍历消息初始化
        if build_depth == 0:
            return
        self.children = []
        for child_ctrl in self.ctrl.GetChildren():
            node = MsgNode(child_ctrl, parent=self, build_depth=build_depth - 1)
            node.init()  # 节点扩展的初始化操作

        # 2 设置每条消息的时间
        last_time = None
        for c in self.children:
            if c.time:
                last_time = c.time
            else:
                c.time = last_time

    def findidx_system_time(self):
        """ 最后条系统时间标记 """
        for idx in range(len(self.children) - 1, -1, -1):
            if self.children[idx].content_type == 'time':
                return idx
        return -1

    def findidx_last_send(self):
        """ 最后条发出去的消息 """
        for idx in range(len(self.children) - 1, -1, -1):
            if self.children[idx].content_type == 'receive':
                return idx
        return -1

    def findidx_last_match(self, old_childrens):
        """
        :param old_chat_box: 旧的消息队列
        :return: 在当前 self.children 中首次匹配到的 old_chat_box 最后几条消息的起始下标
        """
        x, y = old_childrens, self.children
        n, m = len(x), len(y)

        def is_system(a):
            """ 非"撤回"的系统消息 """
            if a.msg_type == 'system' and a.content_type != 'recall':
                return True

        def check_bais(k):
            """ 检查偏移量k是否能匹配 """
            # i, j分别指向"最后一条"数据，然后开始匹配
            i, j = n - 1, m - k - 1
            while min(i, j) > 0:  # 不匹配第0条，第0条太特别
                if is_system(x[i]):
                    i -= 1
                    continue
                if is_system(y[j]):
                    j -= 1
                    continue
                if y[j].is_match(x[i]):
                    i, j = i - 1, j - 1
                    continue
                return False
            return True

        for k in range(m):
            if check_bais(k):
                return m - k - 1

        return -1  # 如果没有匹配，返回 -1

    def check_update(self):
        """ 更新当前消息记录，并返回新收到的消息节点 """
        import copy
        old_childrens = copy.copy(self.children)
        self.build_children()
        idx = self.findidx_last_match(old_childrens)
        return self.children[idx + 1:]


class EditorNode(UiCtrlNode):
    """ 编辑器节点 """


class WeChatMainWnd(UiCtrlNode):

    def __1_结构建构(self):
        pass

    def __init__(self, ctrl=None, parent=None, *, build_depth=12):
        """
        :param ctrl: 当前节点
        :param parent: 父结点
        :param build_depth: 自动构建多少层树节点，默认-1表示构建全部节点
            目前微信发现12层一般够大部分情况使用了
        """
        if ctrl is None:
            ctrl = find_ctrl('WeChatMainWndForPC', '微信')
        super().__init__(ctrl, parent=parent, build_depth=build_depth)

    def build_children(self, build_depth=12, child_node_class=None):
        # 1 构建树结构
        if build_depth == 0:
            return

        # 跳过两级
        self.children = []  # 删除现有的所有子结点
        ctrl = self.ctrl.GetChildren()[-1].GetChildren()[0]
        for child_ctrl in ctrl.GetChildren():
            UiCtrlNode(child_ctrl, parent=self, build_depth=build_depth - 1)

        # 2 预先定义好一些常用的控件
        self.nav = self[0]  # 导航
        self.nav_avatar = self.nav[0]  # 导航_头像
        self.nav_chat = self.nav[1]  # 导航_聊天
        self.nav_contacts = self.nav[2]  # 导航_通讯录
        self.nav_favorites = self.nav[3]  # 导航_收藏
        self.nav_files = self.nav[4]  # 导航_聊天文件
        self.nav_moments = self.nav[5]  # 导航_朋友圈

    def __2_常用控件(self):
        pass

    @property
    def column3(self):
        # 订阅号、聊天都共通
        return self[2][0]

    @property
    def chat_window(self):
        return self.column3[0][0][0]

    @property
    def msg_box(self):
        parent = self.chat_window[1][0][0]
        if not isinstance(parent[0], MsgBoxNode):  # 只初始化一次
            parent.children = (MsgBoxNode(parent[0]), *parent.children[1:])
        return parent[0]

    @property
    def edit_box(self):
        parent = self.chat_window[1][1][1]
        if not isinstance(parent, EditorNode):
            parent.children = (EditorNode(parent[0]), *parent.children[1:])
        return parent[0]

    @property
    def editor(self):
        return self.edit_box[1][0]

    @property
    def subscription_window(self):
        return self.column3[0][0]  # 订阅号窗口

    @property
    def subscription_window_title(self):
        return self.subscription_window[0][0]  # 订阅号窗口标题

    def __3_窗口切换功能(self):
        pass

    def get_chat_with(self):
        """ 当前在跟谁聊天 """
        try:
            return self.editor.Name
        except IndexError:
            return

    def set_chat_with(self, name):
        """ 要跟谁聊天，确保聊天框切换过去 """
        self.nav_chat.Click()

        if self.get_chat_with() == name:
            return

        column2 = self[1]  # 订阅号、聊天都共通
        search_box = column2[0][0][1][0]

        search_box.Click()
        search_box.SendKeys(name)
        time.sleep(2)
        search_box.SendKeys('{Enter}')

    def __4_编辑器(self):
        pass

    def get_editor_content(self):
        """ 获得编辑器正在编辑的内容（涉及到换行数据的时候好像不太准确） """
        return self.editor.GetValuePattern().Value

    def clear_editor_content(self):
        """ 删除编辑区中的所有内容 """
        edit_box = self.edit_box
        edit_box.SendKeys('{Ctrl}a')
        edit_box.SendKeys('{Delete}')

    def write_text(self, text, clear=False, send=False):
        # todo 加at人的功能
        # 1 是否清空旧数据
        if clear:
            self.clear_editor_content()

        # 2 写入新数据
        def should_use_clipboard(text):
            # 简单的策略：如果文本过长或包含特殊字符，则使用剪贴板
            return len(text) > 30 or not text.isprintable() or '{' in text

        edit_box = self.edit_box
        edit_box.Click()
        if should_use_clipboard(text):
            auto.SetClipboardText(text)
            time.sleep(1)
            edit_box.SendKeys('{Ctrl}v')
        else:
            edit_box.SendKeys(text)

        # 3 发送内容
        if send:
            time.sleep(1)
            edit_box.SendKeys('{Enter}')

    def send_text(self, text, clear=False):
        self.write_text(text, clear=clear, send=True)

    def send_files(self, file_paths):
        """
        发送多个文件

        :param list[str] file_paths: 必选参数，为文件的路径
        """
        if copy_files_to_clipboard(file_paths=file_paths):
            edit_box = self.edit_box
            edit_box.Click()
            time.sleep(1)
            edit_box.SendKeys('{Ctrl}v')
            time.sleep(1)
            edit_box.SendKeys('{Enter}')
            time.sleep(1)  # 等待发送动作完成


class WeChatSingletonLock:
    """ 基于 get_autoui_lock 的微信全局唯一单例控制器，确保同一时间仅有一个微信自动化程序在操作 """

    def __init__(self, lock_timeout: Optional[float] = -1):
        # 初始化全局锁
        self.lock = get_autoui_lock(timeout=lock_timeout)
        self.wechat = WeChatMainWnd()

    def __enter__(self):
        # 获取锁并激活微信窗口
        self.lock.acquire()
        self.wechat.activate()
        return self.wechat

    def __exit__(self, exc_type, exc_value, traceback):
        # 释放锁
        self.lock.release()


def wechat_lock_send(user, text=None, files=None, *, timeout=-1):
    """ 使用全局唯一单例锁，确保同一时间仅有一个微信自动化程序在操作 """
    with WeChatSingletonLock(timeout) as we:
        we.set_chat_with(user)
        if text:
            we.send_text(text)
        if files:
            we.send_files(files)


def wechat_handler(message):
    # 获取群名，如果没有指定，不使用此微信发送功能
    user = message.record["extra"].get("wechat_user")
    if user:
        wechat_lock_send(user, message)


if sys.platform == 'win32':
    # 创建专用的微信日志记录器，不绑定默认群名
    wechat_logger = logger.bind(wechat_user='文件传输助手')

    # 添加专用的微信处理器
    wechat_logger.add(wechat_handler,
                      format="{time:YYYY-MM-DD HH:mm:ss} | {level} | {name}:{function}:{line} - {message}")

    """ 往微信发送特殊的日志格式报告
    用法：wechat_logger.bind(wechat_user='文件传输助手').info(message)
    
    或者：
    # 先做好默认群名绑定
    wechat_logger = wechat_logger.bind(wechat_user='考勤管理')
    # 然后就能普通logger用法发送了
    wechat_logger.info('测试')
    """
else:
    # 降级为普通logger
    wechat_logger = logger


def __4_wx():
    """ 别人的原版实现 """


class WxOperation:
    """
    微信群发消息的类，提供了与微信应用交互的方法集，用于发送消息，管理联系人列表等功能。

    Attributes:
    ----------
    wx_window: auto.WindowControl
        微信控制窗口
    input_edit: wx_window.EditControl
        聊天界面输入框编辑控制窗口

    Methods:
    -------
    goto_chat_box(name):
        跳转到 指定好友窗口
    __send_text(*msgs):
        发送文本。
    __send_file(*filepath):
        发送文件
    get_friend_list(tag, num):
        可指定tag，获取好友num页的好友数量
    send_msg(name, msgs, file_paths=None, add_remark_name=False, at_everyone=False,
            text_interval=0.05, file_interval=0.5) -> None:
        向指定的好友或群聊发送消息和文件。支持同时发送文本和文件。
    """

    def __init__(self):
        self.wx_window = None
        self.input_edit = None
        self.wx_window: auto.WindowControl
        self.input_edit: auto.EditControl
        auto.SetGlobalSearchTimeout(Interval.BASE_INTERVAL)
        self.visible_flag: bool = False

    def locate_wechat_window(self):
        if not self.visible_flag:
            wake_up_window(class_name=WeChatConfig.WINDOW_CLASSNAME, name=WeChatConfig.WINDOW_NAME)
            self.wx_window = auto.WindowControl(Name=WeChatConfig.WINDOW_NAME, ClassName=WeChatConfig.WINDOW_CLASSNAME)
            if not self.wx_window.Exists(Interval.MAX_SEARCH_SECOND,
                                         searchIntervalSeconds=Interval.MAX_SEARCH_INTERVAL):
                raise Exception('微信似乎并没有登录!')
            self.input_edit = self.wx_window.EditControl()
            self.visible_flag = bool(self.visible_flag)
        # 微信窗口置顶
        self.wx_window.SetTopmost(isTopmost=True)

    def match_nickname(self, name):
        """获取当前面板的好友昵称"""
        self.input_edit = self.wx_window.EditControl(Name=name)
        if self.input_edit.Exists(Interval.MAX_SEARCH_SECOND, searchIntervalSeconds=Interval.MAX_SEARCH_INTERVAL):
            return self.input_edit
        return False

    def goto_chat_box(self, name: str) -> bool:
        """
        跳转到指定 name好友的聊天窗口。

        Args:
            name(str): 必选参数，好友名称

        Returns:
            None
        """
        if ctrl := self.match_nickname(name):
            ctrl.SetFocus()
            return ctrl

        assert name, "无法跳转到名字为空的聊天窗口"
        self.wx_window.SendKeys(text='{Ctrl}F', waitTime=Interval.BASE_INTERVAL)
        self.wx_window.SendKeys(text='{Ctrl}A', waitTime=Interval.BASE_INTERVAL)
        self.wx_window.SendKey(key=auto.SpecialKeyNames['DELETE'])
        auto.SetClipboardText(text=name)
        time.sleep(Interval.BASE_INTERVAL)
        self.wx_window.SendKeys(text='{Ctrl}V', waitTime=Interval.BASE_INTERVAL)
        # 若有匹配结果，第一个元素的类型为PaneControl
        search_nodes = self.wx_window.ListControl(foundIndex=2).GetChildren()
        if not isinstance(search_nodes.pop(0), auto.PaneControl):
            self.wx_window.SendKeys(text='{Esc}', waitTime=Interval.BASE_INTERVAL)
            raise ValueError("昵称不匹配")
        # 只考虑全匹配, 不考虑好友昵称重名, 不考虑好友昵称与群聊重名
        if search_nodes[0].Name == name:
            self.wx_window.SendKey(key=auto.SpecialKeyNames['ENTER'], waitTime=Interval.BASE_INTERVAL)
            time.sleep(Interval.BASE_INTERVAL)
            return True
        # 无匹配用户, 取消搜索框
        self.wx_window.SendKeys(text='{Esc}', waitTime=Interval.BASE_INTERVAL)
        return False

    def at_at_everyone(self, group_chat_name: str):
        """
        @全部人的操作
        Args:
            group_chat_name(str): 群聊名称

        """
        # 个人 定位 聊天框，取 foundIndex=2，因为左侧聊天List 也可以匹配到foundIndex=1
        # 群聊 定位 聊天框 需要带上群人数，故会匹配失败，所以匹配失败的就是群聊
        result = self.wx_window.TextControl(Name=group_chat_name, foundIndex=2)
        # 只要匹配不上，说明这是个群聊窗口
        if not result.Exists(Interval.MAX_SEARCH_SECOND, searchIntervalSeconds=Interval.MAX_SEARCH_INTERVAL):
            # 寻找是否有 @所有人 的选项
            self.input_edit.SendKeys(text='{Shift}2', waitTime=Interval.BASE_INTERVAL)
            everyone = self.wx_window.ListItemControl(Name='所有人')
            if not everyone.Exists(Interval.MAX_SEARCH_SECOND, searchIntervalSeconds=Interval.MAX_SEARCH_INTERVAL):
                self.input_edit.SendKeys(text='{Ctrl}A', waitTime=Interval.BASE_INTERVAL)
                self.input_edit.SendKeys(text='{Delete}', waitTime=Interval.BASE_INTERVAL)
                return
            self.input_edit.SendKeys(text='{Up}', waitTime=Interval.BASE_INTERVAL)
            self.input_edit.SendKeys(text='{Enter}', waitTime=Interval.BASE_INTERVAL)
            self.input_edit.SendKeys(text='{Enter}', waitTime=Interval.BASE_INTERVAL)

    def __send_text(self, *msgs, wait_time, send_shortcut) -> None:
        """
        发送文本.

        Args:
            input_name(str): 必选参数, 为输入框
            *msgs(str): 必选参数，为发送的文本
            wait_time(float): 必选参数，为动态等待时间
            send_shortcut(str): 必选参数，为发送快捷键

        Returns:
            None
        """

        def should_use_clipboard(text: str):
            # 简单的策略：如果文本过长或包含特殊字符，则使用剪贴板
            return len(text) > 30 or not text.isprintable()

        for msg in msgs:
            assert msg, "发送的文本内容为空"
            self.input_edit.SendKeys(text='{Ctrl}a', waitTime=wait_time)
            self.input_edit.SendKey(key=auto.SpecialKeyNames['DELETE'], waitTime=wait_time)
            self.input_edit.SendKeys(text='{Ctrl}a', waitTime=wait_time)
            self.input_edit.SendKey(key=auto.SpecialKeyNames['DELETE'], waitTime=wait_time)

            if should_use_clipboard(msg):
                auto.SetClipboardText(text=msg)
                time.sleep(wait_time * 2.5)
                self.input_edit.SendKeys(text='{Ctrl}v', waitTime=wait_time * 2)
            else:
                self.input_edit.SendKeys(text=msg, waitTime=wait_time * 2)

            # 设置到剪切板再黏贴到输入框
            self.wx_window.SendKeys(text=f'{send_shortcut}', waitTime=wait_time * 2)

    def __send_file(self, *file_paths, wait_time, send_shortcut) -> None:
        """
        发送文件.

        Args:
            *file_paths(str): 必选参数，为文件的路径
            wait_time(float): 必选参数，为动态等待时间
            send_shortcut(str): 必选参数，为发送快捷键

        Returns:
            None
        """
        # 复制文件到剪切板
        if copy_files_to_clipboard(file_paths=file_paths):
            # 粘贴到输入框
            self.input_edit.SendKeys(text='{Ctrl}V', waitTime=wait_time)
            # 按下回车键
            self.wx_window.SendKeys(text=f'{send_shortcut}', waitTime=wait_time / 2)

            time.sleep(wait_time)  # 等待发送动作完成

    def get_friend_list(self, tag: str = None) -> list:
        """
        获取微信好友名称.

        Args:
            tag(str): 可选参数，如不指定，则获取所有好友

        Returns:
            list
        """
        # 定位到微信窗口
        self.locate_wechat_window()
        # 取消微信窗口置顶
        self.wx_window.SetTopmost(isTopmost=False)
        # 点击 通讯录管理
        self.wx_window.ButtonControl(Name="通讯录").Click(simulateMove=False)
        self.wx_window.ListControl(Name="联系人").ButtonControl(Name="通讯录管理").Click(simulateMove=False)
        # 切换到通讯录管理，相当于切换到弹出来的页面
        contacts_window = auto.GetForegroundControl()
        contacts_window.ButtonControl(Name='最大化').Click(simulateMove=False)

        if tag:
            try:
                contacts_window.ButtonControl(Name="标签").Click(simulateMove=False)
                contacts_window.PaneControl(Name=tag).Click(simulateMove=False)
                time.sleep(Interval.BASE_INTERVAL * 2)
            except LookupError:
                contacts_window.SendKey(auto.SpecialKeyNames['ESC'])
                raise LookupError(f'找不到 {tag} 标签')

        name_list = list()
        last_names = None
        while True:
            # TODO 修改成使用 foundIndex 的方式
            try:
                nodes = contacts_window.ListControl(foundIndex=2).GetChildren()
            except LookupError:
                nodes = contacts_window.ListControl().GetChildren()
            cur_names = [node.TextControl().Name for node in nodes]

            # 如果滚动前后名单未变，认为到达底部
            if cur_names == last_names:
                break
            last_names = cur_names
            # 处理当前页的名单
            for node in nodes:
                # TODO 如果有需要, 可以处理成导出为两列的csv格式
                nick_name = node.TextControl().Name  # 用户名
                remark_name = node.ButtonControl(foundIndex=2).Name  # 用户备注名，索引1会错位，索引2是备注名，索引3是标签名
                name_list.append(remark_name if remark_name else nick_name)
            # 向下滚动页面
            contacts_window.WheelDown(wheelTimes=8, waitTime=Interval.BASE_INTERVAL / 2)
        # 结束时候关闭 "通讯录管理" 窗口
        contacts_window.SendKey(auto.SpecialKeyNames['ESC'])
        # 简单去重，但是存在误判（如果存在同名的好友), 保持获取时候的顺序
        return list(dict.fromkeys(name_list))

    def get_group_chat_list(self) -> list:
        """获取群聊通讯录中的用户名称"""
        name_list = list()
        auto.ButtonControl(Name='聊天信息').Click()
        time.sleep(0.5)
        chat_members_win = self.wx_window.ListControl(Name='聊天成员')
        if not chat_members_win.Exists():
            return list()
        self.wx_window.ButtonControl(Name='查看更多').Click()
        for item in chat_members_win.GetChildren():
            name_list.append(item.ButtonControl().Name)
        return name_list

    def send_msg(self, name, msgs=None, file_paths=None, add_remark_name=False, at_everyone=False,
                 text_interval=Interval.SEND_TEXT_INTERVAL, file_interval=Interval.SEND_FILE_INTERVAL,
                 send_shortcut='{Enter}') -> None:
        """
        发送消息，可同时发送文本和文件（至少选一项

        Args:
            name(str):必选参数，接收消息的好友名称, 可以单发
            msgs(Iterable[str], Optional): 可选参数，发送的文本消息
            file_paths(Iterable[str], Optional):可选参数，发送的文件路径
            add_remark_name(bool): 可选参数，是否添加备注名称发送
            at_everyone(bool): 可选参数，是否@全部人
            text_interval(float): 可选参数，默认为0.05
            file_interval(float): 可选参数，默认为0.5
            send_shortcut(str): 可选参数，默认为 Enter

        Raises:
            ValueError: 如果用户名为空或发送的消息和文件同时为空时抛出异常
            TypeError: 如果发送的文本消息或文件路径类型不是列表或元组时抛出异常
        """
        # 定位到微信窗口
        self.locate_wechat_window()

        if not name:
            raise ValueError("用户名不能为空")

        if not any([msgs, file_paths]):
            raise ValueError("发送的消息和文件不可同时为空")

        if msgs and not isinstance(msgs, Iterable):
            raise TypeError("发送的文本消息必须是可迭代的")

        if file_paths and not isinstance(file_paths, Iterable):
            raise TypeError("发送的文件路径必须是可迭代的")

        # 如果当前面板已经是需发送好友, 则无需再次搜索跳转
        if not self.match_nickname(name=name):
            if not self.goto_chat_box(name=name):
                raise NameError('昵称不匹配')

        # 设置输入框为当前焦点
        self.input_edit = self.wx_window.EditControl(Name=name)
        self.input_edit.SetFocus()

        # @所有人
        if at_everyone:
            auto.SetGlobalSearchTimeout(Interval.BASE_INTERVAL)
            self.at_at_everyone(group_chat_name=name)
            auto.SetGlobalSearchTimeout(Interval.BASE_INTERVAL * 25)

        # TODO 添加备注可以多做一个选项，添加到每条消息的前面，如xxx，早上好
        if msgs and add_remark_name:
            new_msgs = deepcopy(list(msgs))
            new_msgs.insert(0, name)
            self.__send_text(*new_msgs, wait_time=text_interval, send_shortcut=send_shortcut)
        elif msgs:
            self.__send_text(*msgs, wait_time=text_interval, send_shortcut=send_shortcut)
        if file_paths:
            self.__send_file(*file_paths, wait_time=file_interval, send_shortcut=send_shortcut)

        # 取消微信窗口置顶
        self.wx_window.SetTopmost(isTopmost=False)


if __name__ == '__main__':
    # wx = WxOperation()
    # data = wx.get_friend_list('无标签')
    # print(data)
    # print(len(data))
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/05/29

from pyxllib.prog.pupil import check_install_package

# cachetools，通用缓存工具，适用范围更广
check_install_package('cachetools')
# cached-property，类属性特用工具，相对比较简洁些
check_install_package('cached_property', 'cached-property')

# 对于普通函数，一般用lru_cache即可
from functools import lru_cache
import threading

from cachetools import cached, LRUCache, TTLCache

# 官方文档：https://pypi.org/project/cached-property/
from cached_property import cached_property
from cached_property import threaded_cached_property  # 线程安全
from cached_property import cached_property_with_ttl  # 限时缓存，单位秒
from cached_property import threaded_cached_property_with_ttl  # 线程 + 限时


# todo 240609周日21:19 https://github.com/awolverp/cachebox，据说这个缓存库速度更快的多

# 进一步封装的更通用、自己常用的装饰器

def xlcache(maxsize=128, *, ttl=None, lock=None, property=False):
    """ 那些工具接口太复杂难记，自己封装一个统一的工具

    就是一个装饰器，最大缓存多少项，然后是否要开多线程安全，是否要设置限时重置，是否是作为类成员属性修饰

    :param property: 是否作为类成员属性修饰，不过一般不建议通过这里设置，
        而是外部再加一层@property，不然IDE会识别不了这是一个property，影响开发

    """
    def decorator(func):
        if property:
            if ttl is not None:
                if lock:
                    # 使用带有时间限制和线程安全的缓存属性
                    return threaded_cached_property_with_ttl(ttl)(func)
                else:
                    # 使用带有时间限制但非线程安全的缓存属性
                    return cached_property_with_ttl(ttl)(func)
            else:
                if lock:
                    # 使用线程安全的缓存属性
                    return threaded_cached_property(func)
                else:
                    # 使用普通的缓存属性
                    return cached_property(func)
        else:
            lock2 = threading.RLock() if lock is True else lock
            if ttl is None:
                return cached(LRUCache(maxsize), lock=lock2)(func)
            else:
                cache = TTLCache(maxsize=maxsize, ttl=ttl)
                return cached(cache, lock=lock2)(func)

    return decorator
